;; NAME-NOT-IN.SCM

;; NAME-NOT-IN returns the first version of the "root" of the symbol 
;; TEMPLATE with a non-negative "subscript" that is not in NAME-SET.
;; E.g. 
;; 
;; > (name-not-in 'a '(b c d))
;; a_0
;; 
;; > (name-not-in 'a '(a b c d))
;; a_0
;; 
;; > (name-not-in 'a '(b c a_0 d))
;; a_1
;; 
;; > (name-not-in 'a '(a_0 a_4 a_2 b a_1 c d))
;; a_3
;; 
;; > (name-not-in 'a_0 '(a_0 a_4 a_2 b a_1 c d))
;; a_3
;; 
;; > (name-not-in 'a_3 '(a_0 a_4 a_2 b a_1 c d))
;; a_3

;; The implementation below uses several functions --
;; pre-defined functions from the Scheme Standard Library: 
;;    string->symbol, string-append, symbol->string, 
;;    number->string, substring, string-length, string-ref, 
;;    char>=?, and char<=?,
;; help functions defined in this file:
;;    name-search, root, subscript, root-loop, and digit?,
;; help functions defined in the file "set-list-nodups.scm":
;;    set-member?

(load "Scheme-utilities/set-list-nodups.scm")

(define name-not-in
  (lambda (template name-set)
    (name-search (root template) 0 name-set)))

(define name-search 
  (lambda (symbol int name-set)
    (let ((subscripted (subscript symbol int)))
      (if (not (set-member? subscripted name-set))
        subscripted
        (name-search symbol (+ int 1) name-set)))))

(define subscript
  (lambda (symbol int) 
    (string->symbol 
     (string-append
      (symbol->string symbol)
      "_"
      (number->string int)))))

;; Returns the "root" of a name -- i.e., the part before
;; any final underscore followed by a number.
(define root
  (lambda (symbol)
    (let ((str (symbol->string symbol)))
      (string->symbol (substring str
				 0
				 (root-loop str (- (string-length str) 1) #f))))))

;; Returns index of final underscore in string if it is followed
;; by a non-empty number of digits. In all other cases, returns
;; length of string. 
(define root-loop
  (lambda (str index seen-digit?)
    (let ((char (string-ref str index)))
      (cond ((= index 0) (string-length str))
            ((char=? char #\_)
	     (if seen-digit? index (string-length str)))
            ((digit? char)
	     (root-loop str (- index 1) #t))
            (else (string-length str))))))

      
(define digit? 
  (lambda (char)
    (or (char>=? char #\0) (char<=? char #\9))))
