;;; SET-LIST-NODUPS.SCM
;;; Implementation of set ADT in terms of tagged lists without duplicates

; (load "../Scheme-utilities/list-ops.scm")

(define set-tag (list '*i-am-a-set*))

(define set?
  (lambda (s) 
    (and (pair? s)
	 (eq? (car s) set-tag))))

(define set-make
  (lambda (elts) 
    (cons set-tag elts)))

(define set-contents
  (lambda (set)
    (if (set? set)
	(cdr set)
	(throw 'set-contents:not-a-set set))))

(define set-empty
  (lambda () 
    (set-make '())))

(define set-empty?
  (lambda (s) 
    (if (not (set? s))
	(throw 'set-empty?:not-a-set s)
	(null? (set-contents s)))))

(define set-singleton
  (lambda (x) 
    (set-make (list x))))

(define set-member? 
  (lambda (x s)
    (if (not (set? s))
	(throw 'set-member?:not-a-set s)
	(member x (set-contents s)))))

(define set-insert
  (lambda (x s)
    (if (not (set? s))
	(throw 'set-insert:not-a-set s)
	(if (set-member? x s)
	    s
	    (set-make (cons x (set-contents s)))))))

(define set-choose
  (lambda (s) 
    (if (not (set? s))
	(throw 'set-choose:not-a-set s)
	(if (null? (set-contents s))
	    (throw 'set-choose:empty-set s)
	    (car (set-contents s))))))

(define set-union
  (lambda (s1 s2)
    (cond 
     ((not (set? s1)) (throw 'set-union:not-a-set s1))
     ((not (set? s2)) (throw 'set-union:not-a-set s2))
     (else
      (set-make
       (let ((c2 (set-contents s2)))
	 (let union-all ((c1 (set-contents s1)))
	   (cond ((null? c1) c2)
		 ((member (car c1) c2) (union-all (cdr c1)))
		 (else (cons (car c1) (union-all (cdr c1))))))))))))

(define set-intersection
  (lambda (s1 s2)
    (cond 
     ((not (set? s1)) (throw 'set-intersection:not-a-set s1))
     ((not (set? s2)) (throw 'set-intersection:not-a-set s2))
     (else
      (set-make
       (let ((c2 (set-contents s2)))
	 (let isect-all ((c1 (set-contents s1)))
	   (cond ((null? c1) '())
		 ((member (car c1) c2) 
		  (cons (car c1) (isect-all (cdr c1))))
		 (else (isect-all (cdr c1)))))))))))

(define set-difference
  (lambda (s1 s2)
    (cond 
     ((not (set? s1)) (throw 'set-difference:not-a-set s1))
     ((not (set? s2)) (throw 'set-difference:not-a-set s2))
     (else
      (set-make
       (let ((c2 (set-contents s2)))
	 (let diff-all ((c1 (set-contents s1)))
	   (cond ((null? c1) '())
		 ((member (car c1) c2) 
		  (diff-all (cdr c1)))
		 (else (cons (car c1) (diff-all (cdr c1))))))))))))

(define set-subset?
  (lambda (s1 s2)
    (cond 
     ((not (set? s1)) (throw 'set-subset?:not-a-set s1))
     ((not (set? s2)) (throw 'set-subset?:not-a-set s2))
     (else
      (set-forall? (lambda (x1) (set-member? x1 s2)) s1)))))

(define set-equal?
  (lambda (s1 s2)
    (cond 
     ((not (set? s1)) (throw 'set-equal?:not-a-set s1))
     ((not (set? s2)) (throw 'set-equal?:not-a-set s2))
     (else
      (and (set-subset? s1 s2)
	   (set-subset? s2 s1))))))

(define list->set
  (lambda (lst)
    (if (null? lst)
	(set-empty)
	(set-insert (car lst)
		    (list->set (cdr lst))))))

(define set->list
  (lambda (s) 
    (if (not (set? s))
	(throw 'set->list:not-a-set s)
	(set-contents s))))

(define set-map
  (lambda (f s)
    (if (not (set? s))
	(throw 'set-map:not-a-set s)
	(set-make
	 (map f (set-contents s))))))

(define set-filter
  (lambda (p s)
    (if (not (set? s))
	(throw 'set-filter:not-a-set s)
	(set-make
	 (filter p (set-contents s))))))

(define set-fold
  ;; Should use a binop that is order indepedent,
  ;; since order of elements in fold cannot be predicted.
  (lambda (binop null s)
    (if (not (set? s))
	(throw 'set-fold:not-a-set s)
	(foldr binop null (set-contents s)))))

(define set-union-map 
  (lambda (f s)
    (if (not (set? s))
	(throw 'set-union-map:not-a-set s)
	(foldr set-union (set-empty) (map f (set-contents s))))))

(define set-forall?
  (lambda (f s)
    (if (not (set? s))
	(throw 'set-forall?:not-a-set s)
	(forall? f (set-contents s)))))
