;;; TEST.SCM
;;; General testing functions
;;; History: 
;;; + [lyn, 2/10/01] Created


(define test 
  (lambda (fcn suite)
    (for-each (lambda (tester) (tester fcn))
	      suite)))

(define test-separated
  (lambda (fcn suite)
    (for-each (lambda (tester) (line) (tester fcn))
	      suite)))

(define test-benchmark
  (lambda (fcn)
    (lambda (benchmark)
      (let ((args (first benchmark))
	    (tester (second benchmark)))
	(tester (apply fcn args))))))

;; Test a given function based on its input/ouput behavior 
;; relative to given input and expected output.
(define equality-tester
  (lambda (namer) 
    (lambda (args expected)
      (lambda (fcn)
	(begin
	  (display (apply namer args))
	  (let ((actual (apply fcn args)))
	    (if (equal? actual expected)
		(display " OK!\n")
		(begin (display "\n\n********** ERROR **********\n")
		       (display (apply namer args))
		       (display "\nThe actual result:\n")
		       (display actual)
		       (display "\ndiffers from the expected result:\n")
		       (display expected)
		       (display "\n***************************\n\n")))))))))


;; Like equality-tester, but gives control over how success
;; is indicated via success function. 
(define equality-success-tester
  (lambda (namer succeed)
    (lambda (args expected)
      (lambda (fcn)
	(begin
	  (display (apply namer args))
	  (let ((actual (apply fcn args)))
	    (if (equal? actual expected)
		(succeed actual)
		(begin (display "\n\n********** ERROR **********\n")
		       (display (apply namer args))
		       (display "\nThe actual result:\n")
		       (display actual)
		       (display "\ndiffers from the expected result:\n")
		       (display expected)
		       (display "\n***************************\n\n")))))))))

;; Test a given function based on its input/ouput behavior relative 
;; to a standard implementxation of the function on given input.
(define comparison-tester
  (lambda (namer wrapper standard)
    (lambda (args)
      (lambda (fcn)
	(begin
	  (display (apply namer args))
	  (let ((actual (apply (wrapper fcn) args))
		(expected (apply (wrapper standard) args)))
	    (if (equal? actual expected)
		(display " OK!\n")
		(begin (display "\n\n********** ERROR **********\n")
		       (display (apply namer args))
		       (display "\nThe actual result:\n")
		       (display actual)
		       (display "\ndiffers from the expected result:\n")
		       (display expected)
		       (display "\n***************************\n\n")))))))))

(define standard-namer
  (lambda (name)
    (lambda args
      (cons name args))))

(define map-apply
  (lambda (f lst)
    (map (lambda (args)
	   (apply f args))
	 lst)))

(define line
  (lambda () 
    (display "----------------------------------------------------------------------\n")
    ))

(define test-all-testers
  (lambda (testers)
    (begin 
      (line)
      (for-each 
       (lambda (tester)
	 (begin (tester) (line)))
       testers))))


;;--------------------------------------------------------------------------
;; The following are used by the various language testers.

(define run-test-cases
  (lambda (run-string error-sym)
    (define has-inputs? 
      (lambda (test-case)
	(= (length test-case) 4)))
    (define get-inputs
      (lambda (test-case)
	(third test-case)))
    (define get-expected-result
      (lambda (test-case)
	(if (has-inputs? test-case)
	    (fourth test-case)
	    (third test-case))))
    (define run-test-case
      (lambda (run)
	(lambda (test-case)
	  (let* ((program-name (first test-case))
		 (program (second test-case))
		 (inputs (if (has-inputs? test-case) (get-inputs test-case) 'none))
		 (expected-result (get-expected-result test-case))
		 (prefix (if (has-inputs? test-case)
			     (build-string "\n" run-string " " program-name " on " 
					   inputs " gives ")
			     (build-string "\n" run-string program-name " gives ")))
		 )
	    (begin
	      ;; Display prefix right away so there's
	      ;; some feedback if computation loops.
	      (display prefix)
	      (let ((result (catch (lambda () 
				     (if (has-inputs? test-case)
					 (run program inputs)
					 (run program)))
				   (lambda (tag value) (list error-sym tag value)))))
		(if (equal? result expected-result)
		    (display (build-string result ". OK!"))
		    (let ((suffix (build-string result
						" ***ERROR!*** Expected "
						expected-result)))
		      (found-error prefix suffix)))))))))

    (lambda (run suite)
      (begin
	(for-each (run-test-case run) suite)
	(newline)
	(display "Done.")
	(newline)))
    ))

;;; The following uses side effects (set!) to count the
;;; number of errors. We will study set! later in the semester.
(define *errors* '())

(define found-error 
  (lambda (prefix suffix)
    (begin 
      (set! *errors* (cons (string-append prefix suffix) *errors*))
      (display suffix)
      )))


(define error-status
  (lambda () 
    (begin
      (line)
      (display "***ERROR SUMMARY***")
      (for-each display *errors*)
      (newline)
      (line)
      (display "Number of errors encountered: ")
      (display (length *errors*))
      (set! *errors* '()) ; reset for next time
      (newline)
      (line))))

;; Turn an arbitrary length list of arguments
;; into a single string
(define build-string
  (lambda args
    (with-output-to-string
      (lambda ()
	(for-each display args)))))

;; This is a hack. 
;; Should really change generic tester
;; to handle lists of inputs/outputs!
(define multi-test-suite->single-test-suite
  (lambda (multi-test-suite)  
    (foldr append '()
	   (map (lambda (test-case)
		  (apply
		   (lambda (name pgm inputs/outputs)
		     (map (lambda (input/output)
			    (list name 
				  pgm 
				  (first input/output) 
				  (second input/output)))
			  inputs/outputs))
		   test-case))
		multi-test-suite))))


